import path from 'path';
import { CosmosConfig } from '../../config';
import slash from 'slash';
import { findUserModulePaths } from './findUserModulePaths';

// Warning: Renderer config must be serializable!
export function generateUserDepsModule(
  cosmosConfig: CosmosConfig,
  rendererConfig: {}
): string {
  const {
    rootDir,
    fixturesDir,
    fixtureFileSuffix,
    globalImports,
  } = cosmosConfig;

  const { fixturePaths, decoratorPaths } = findUserModulePaths({
    rootDir,
    fixturesDir,
    fixtureFileSuffix,
  });

  return getCompiledTemplate({
    globalImports: genGlobalRequires(globalImports),
    rendererConfig: stringifyRendererConfig(rendererConfig),
    fixtures: genRequireMap(fixturePaths, rootDir),
    decorators: genRequireMap(decoratorPaths, rootDir),
  });
}

function genGlobalRequires(paths: string[]) {
  if (paths.length === 0) {
    return '';
  }

  // Forward slashes are necessary via slash() to avoid double-escaping
  // backslashes on Windows
  return [
    '',
    '// Keeping global imports here is superior to making them bundle entry points',
    '// because this way they become hot-reloadable',
    ...paths.map(importPath => `require('${slash(importPath)}');`),
    '',
  ].join(`\n`);
}

function genRequireMap(paths: string[], rootDir: string) {
  if (paths.length === 0) {
    return '{}';
  }

  const requireRows = paths.map(p => {
    const relPath = slash(path.relative(rootDir, p));
    return `
  '${relPath}': require('${p}').default`;
  });

  return `{${requireRows.join(', ')}\n}`;
}

function stringifyRendererConfig(rendererConfig: {}) {
  return JSON.stringify(rendererConfig, null, 2);
}

export type TemplateArgs = {
  globalImports: string;
  rendererConfig: string;
  fixtures: string;
  decorators: string;
};

function getCompiledTemplate({
  globalImports,
  rendererConfig,
  fixtures,
  decorators,
}: TemplateArgs) {
  return `// This file is automatically generated by Cosmos
${globalImports}
export const rendererConfig = ${rendererConfig};
export const fixtures = ${fixtures};
export const decorators = ${decorators};
`;
}
